var toStringFunction = Function.prototype.toString;
var create = Object.create, defineProperty = Object.defineProperty, getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor, getOwnPropertyNames = Object.getOwnPropertyNames, getOwnPropertySymbols = Object.getOwnPropertySymbols;
var _a$1 = Object.prototype, hasOwnProperty = _a$1.hasOwnProperty, propertyIsEnumerable = _a$1.propertyIsEnumerable;
var SYMBOL_PROPERTIES = typeof getOwnPropertySymbols === 'function';
var LegacyCache = /** @class */ (function () {
    function LegacyCache() {
        this._keys = [];
        this._values = [];
    }
    LegacyCache.prototype.has = function (key) {
        return !!~this._keys.indexOf(key);
    };
    LegacyCache.prototype.get = function (key) {
        return this._values[this._keys.indexOf(key)];
    };
    LegacyCache.prototype.set = function (key, value) {
        this._keys.push(key);
        this._values.push(value);
    };
    return LegacyCache;
}());
function createCacheLegacy() {
    return new LegacyCache();
}
function createCacheModern() {
    return new WeakMap();
}
/**
 * Get a new cache object to prevent circular references.
 */
var createCache = typeof WeakMap !== 'undefined' ? createCacheModern : createCacheLegacy;
function getArrayCloneLoose(array, prototype, handleCopy, cache) {
    var clone = new prototype.constructor();
    cache.set(array, clone);
    for (var index = 0, length_1 = array.length; index < length_1; ++index) {
        clone[index] = handleCopy(array[index], cache);
    }
    return clone;
}
/**
 * Get an empty version of the object with the same prototype it has.
 */
function getCleanClone(object, prototype) {
    if (!prototype) {
        return create(null);
    }
    var Constructor = prototype.constructor;
    if (Constructor === Object) {
        return prototype === Object.prototype ? {} : create(prototype);
    }
    if (~toStringFunction.call(Constructor).indexOf('[native code]')) {
        try {
            return new Constructor();
        }
        catch (_a) { }
    }
    return create(prototype);
}
/**
 * Get a copy of the object based on loose rules, meaning all enumerable keys
 * and symbols are copied, but property descriptors are not considered.
 */
function getObjectCloneLoose(object, prototype, handleCopy, cache) {
    var clone = getCleanClone(object, prototype);
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    for (var key in object) {
        if (hasOwnProperty.call(object, key)) {
            clone[key] = handleCopy(object[key], cache);
        }
    }
    if (SYMBOL_PROPERTIES) {
        var symbols = getOwnPropertySymbols(object);
        for (var index = 0, length_2 = symbols.length, symbol = void 0; index < length_2; ++index) {
            symbol = symbols[index];
            if (propertyIsEnumerable.call(object, symbol)) {
                clone[symbol] = handleCopy(object[symbol], cache);
            }
        }
    }
    return clone;
}
function getStrictPropertiesModern(object) {
    return getOwnPropertyNames(object).concat(getOwnPropertySymbols(object));
}
var getStrictProperties = SYMBOL_PROPERTIES
    ? getStrictPropertiesModern
    : getOwnPropertyNames;
/**
 * Get a copy of the object based on strict rules, meaning all keys and symbols
 * are copied based on the original property descriptors.
 */
function getObjectCloneStrict(object, prototype, handleCopy, cache) {
    var clone = getCleanClone(object, prototype);
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    var properties = getStrictProperties(object);
    for (var index = 0, length_3 = properties.length, property = void 0, descriptor = void 0; index < length_3; ++index) {
        property = properties[index];
        if (property !== 'callee' && property !== 'caller') {
            descriptor = getOwnPropertyDescriptor(object, property);
            if (descriptor) {
                // Only clone the value if actually a value, not a getter / setter.
                if (!descriptor.get && !descriptor.set) {
                    descriptor.value = handleCopy(object[property], cache);
                }
                try {
                    defineProperty(clone, property, descriptor);
                }
                catch (error) {
                    // Tee above can fail on node in edge cases, so fall back to the loose assignment.
                    clone[property] = descriptor.value;
                }
            }
            else {
                // In extra edge cases where the property descriptor cannot be retrived, fall back to
                // the loose assignment.
                clone[property] = handleCopy(object[property], cache);
            }
        }
    }
    return clone;
}
function getRegExpFlagsLegacy(regExp) {
    var flags = '';
    if (regExp.global) {
        flags += 'g';
    }
    if (regExp.ignoreCase) {
        flags += 'i';
    }
    if (regExp.multiline) {
        flags += 'm';
    }
    if (regExp.unicode) {
        flags += 'u';
    }
    if (regExp.sticky) {
        flags += 'y';
    }
    return flags;
}
function getRegExpFlagsModern(regExp) {
    return regExp.flags;
}
/**
 * Get the flags to apply to the copied regexp.
 */
var getRegExpFlags = /test/g.flags === 'g' ? getRegExpFlagsModern : getRegExpFlagsLegacy;

var _a, _b;
var isArray = Array.isArray;
var getPrototypeOf = Object.getPrototypeOf;
var toString = Object.prototype.toString;
var ARRAY_BUFFER_OBJECT_CLASSES = (_a = {},
    _a['[object ArrayBuffer]'] = true,
    _a['[object Float32Array]'] = true,
    _a['[object Float64Array]'] = true,
    _a['[object Int8Array]'] = true,
    _a['[object Int16Array]'] = true,
    _a['[object Int32Array]'] = true,
    _a['[object Uint8Array]'] = true,
    _a['[object Uint8ClampedArray]'] = true,
    _a['[object Uint16Array]'] = true,
    _a['[object Uint32Array]'] = true,
    _a['[object Uint64Array]'] = true,
    _a);
var UNCOPIABLE_OBJECT_CLASSES = (_b = {},
    _b['[object Error]'] = true,
    _b['[object Promise]'] = true,
    _b['[object WeakMap]'] = true,
    _b['[object WeakSet]'] = true,
    _b);
function performCopy(value, getObjectClone, getArrayClone) {
    function handleCopy(value, cache) {
        if (!value || typeof value !== 'object') {
            return value;
        }
        if (cache.has(value)) {
            return cache.get(value);
        }
        var prototype = value.__proto__ || getPrototypeOf(value);
        var Constructor = prototype && prototype.constructor;
        // plain objects
        if (!Constructor || Constructor === Object) {
            return getObjectClone(value, prototype, handleCopy, cache);
        }
        // arrays
        if (isArray(value)) {
            return getArrayClone(value, prototype, handleCopy, cache);
        }
        var objectClass = toString.call(value);
        // dates
        if (objectClass === '[object Date]') {
            return new Constructor(value.getTime());
        }
        // regexps
        if (objectClass === '[object RegExp]') {
            var clone = new Constructor(value.source, value.flags || getRegExpFlags(value));
            clone.lastIndex = value.lastIndex;
            return clone;
        }
        // maps
        if (objectClass === '[object Map]') {
            var clone = new Constructor(value.entries());
            cache.set(value, clone);
            return clone;
        }
        // sets
        if (objectClass === '[object Set]') {
            var clone = new Constructor(value.values());
            cache.set(value, clone);
            return clone;
        }
        // blobs
        if (objectClass === '[object Blob]') {
            return value.slice(0, value.size, value.type);
        }
        // dataviews
        if (objectClass === '[object DataView]') {
            var clone = new Constructor(value.buffer.slice(0));
            cache.set(value, clone);
            return clone;
        }
        // array buffers
        if (ARRAY_BUFFER_OBJECT_CLASSES[objectClass]) {
            var clone = value.slice(0);
            cache.set(value, clone);
            return clone;
        }
        // if the value cannot / should not be copied deeply, return the reference
        if (
        // promise-like
        typeof value.then === 'function' ||
            // object classes which cannot be introspected for copy
            UNCOPIABLE_OBJECT_CLASSES[objectClass]) {
            return value;
        }
        // assume anything left is a custom constructor
        return getObjectClone(value, prototype, handleCopy, cache);
    }
    return handleCopy(value, createCache());
}
/**
 * Copy an value deeply as much as possible.
 */
function copy(value) {
    return performCopy(value, getObjectCloneLoose, getArrayCloneLoose);
}
/**
 * Copy an value deeply as much as possible, where strict recreation of object properties
 * are maintained. All properties (including non-enumerable ones) are copied with their
 * original property descriptors on both objects and arrays.
 */
function copyStrict(value) {
    return performCopy(value, getObjectCloneStrict, getObjectCloneStrict);
}

export { copy, copyStrict };
//# sourceMappingURL=index.mjs.map
